#include <glib.h>
#include <gtk/gtk.h>
#include <stdlib.h>

#include <X11/SM/SMlib.h>
#include <X11/ICE/ICElib.h>
#include <signal.h>

#include <main.h>

SmcConn     smc_conn = NULL;
IceConn     ice_conn = NULL;
char       *client_id = NULL;
GIOChannel *ice_io_channel = NULL;
int         ice_handler_id;

gboolean    terminate = FALSE;

static void
print_session_information (SmcConn smc_conn)
{
  char *str;

  g_message ("%s: XSMP Version: %d  Revision: %d", G_STRLOC,
           SmcProtocolVersion (smc_conn),
           SmcProtocolRevision (smc_conn));

  str = SmcVendor (smc_conn);
  g_message ("%s: Session manager: %s ", G_STRLOC, str);
  free (str);

  str = SmcRelease (smc_conn);
  g_message ("%s: Release: %s", G_STRLOC, str);
  free (str);
}

static void
session_save_yourself (SmcConn   smc_conn,
                       SmPointer client_data,
                       int       save_type,
                       Bool      shutdown,
                       int       interact_style,
                       Bool      fast)
{
  g_message ("%s: Save yourself", G_STRLOC);

  switch (interact_style)
    {
    case SmInteractStyleNone:
      g_message ("%s: No interaction", G_STRLOC);
      break;

    case SmInteractStyleErrors:
      g_message ("%s: Interaction only under an error condition", G_STRLOC);
      break;

    case SmInteractStyleAny:
      g_message ("%s: Any interaction", G_STRLOC);

/*       SmcInteractRequest (smc_conn); */
/*       SmcInteractDone (smc_conn, False); */
      break;
    }

  switch (save_type)
    {
    case SmSaveLocal:
      g_message ("%s: Save local state", G_STRLOC);
      break;

    case SmSaveGlobal:
      g_message ("%s: Save global state", G_STRLOC);
      break;

    case SmSaveBoth:
      g_message ("%s: Save both global and local state", G_STRLOC);
      break;
    }

  if (shutdown)
    {
      g_message ("%s: Is shutting down", G_STRLOC);
    }

  SmcSaveYourselfDone (smc_conn, True);
}

static void
session_die (SmcConn   conenction,
             SmPointer client_data)
{
  g_message ("%s: Session die", G_STRLOC);
}

static void
session_save_complete (SmcConn   smc_conn,
                       SmPointer client_data)
{
  g_message ("%s: Save complete", G_STRLOC);
}

static void
session_shutdown_cancelled (SmcConn   smc_conn,
                            SmPointer client_data)
{
  g_message ("%s: Shutdown cancelled", G_STRLOC);
}

static gboolean
ice_handler (GIOChannel   *io_channel,
             GIOCondition  condition,
             gpointer      data)
{
  IceConn ice_conn = (IceConn) data;
  
  if (condition == G_IO_IN)
    {
      IceReplyWaitInfo info;
      Bool ready;
      IceProcessMessagesStatus status;

      status = IceProcessMessages (ice_conn, &info, &ready);

      switch (status)
        {
        case IceProcessMessagesSuccess:
          g_message ("%s: ICE message processing ok", G_STRLOC);
          break;

        case IceProcessMessagesIOError:
          g_message ("%s: I/O error occurred while processing ICE messages", G_STRLOC);
          break;

        case IceProcessMessagesConnectionClosed:
          g_message ("%s: ICE connection already closed!", G_STRLOC);
          break;
        }

      if (ready)
        {
          g_message ("%s: Reply is ready", G_STRLOC);
        }
    }

  return TRUE;
}

void
session_start (void)
{
  char error[256] = "";

  SmcCallbacks callbacks = 
    { 
      {session_save_yourself, NULL},
      {session_die, NULL},
      {session_save_complete, NULL},
      {session_shutdown_cancelled, NULL}
    };

  smc_conn = SmcOpenConnection (NULL, NULL, 1, 0, 
                                SmcSaveYourselfProcMask | 
                                SmcDieProcMask |
                                SmcSaveCompleteProcMask | 
                                SmcShutdownCancelledProcMask,
                                &callbacks,
                                NULL, &client_id,
                                sizeof(error), error);

  if (smc_conn)
    {
      g_message ("%s: Connection opened, client id is %s", G_STRLOC, client_id);
      print_session_information (smc_conn);

      ice_conn = SmcGetIceConnection (smc_conn);
      ice_io_channel = g_io_channel_unix_new (IceConnectionNumber (ice_conn));

      ice_handler_id = g_io_add_watch (ice_io_channel, G_IO_IN, ice_handler, ice_conn);
    }
  else
    {
      g_message ("%s:\t\t\tUnable to open SMlib connection: %s", __FILE__, error);
    }
}

void
session_end (void)
{
  SmcCloseStatus status;
  GError *error = NULL;

  if (!smc_conn)
    return;

  g_source_remove (ice_handler_id);
  g_io_channel_unref (ice_io_channel);

  status = SmcCloseConnection (smc_conn, 0, NULL);

  if (error)
    {
      g_message ("%s: Error shutting down ICE IO channel: %s", G_STRLOC,
               error->message);
      g_error_free (error);
    }

  free (client_id);

  switch (status)
    {
    case SmcClosedNow: 
      g_message ("%s: Connection closed now", G_STRLOC);
      break;

    case SmcClosedASAP:
      g_message ("%s: Connection closed ASAP", G_STRLOC);
      break;

    case SmcConnectionInUse:
      g_message ("%s: Connection in use", G_STRLOC);
      break;

    default:
      g_message ("%s: ??", G_STRLOC);
    }
}

#if 0
void
sigterm_handler (int signum)
{
  terminate = TRUE;
}

void
signal_handlers_setup (void)
{
  struct sigaction action;

  action.sa_handler = sigterm_handler;
  action.sa_flags = SA_RESTART;
  sigemptyset (&action.sa_mask);

  sigaction (SIGTERM, &action, NULL);
}
#endif
